/*
 * All of the documentation and software included in the
 * Alchemy Software is copyrighted by Stanley Kok, Parag
 * Singla, Matthew Richardson, Pedro Domingos, Marc
 * Sumner and Hoifung Poon.
 * 
 * Copyright [2004-06] Stanley Kok, Parag Singla, Matthew
 * Richardson, Pedro Domingos, Marc Sumner and Hoifung
 * Poon. All rights reserved.
 * 
 * Contact: Pedro Domingos, University of Washington
 * (pedrod@cs.washington.edu).
 * 
 * Redistribution and use in source and binary forms, with
 * or without modification, are permitted provided that
 * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above
 * copyright notice, this list of conditions and the
 * following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the
 * above copyright notice, this list of conditions and the
 * following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use
 * of this software must display the following
 * acknowledgment: "This product includes software
 * developed by Stanley Kok, Parag Singla, Matthew
 * Richardson, Pedro Domingos, Marc Sumner and Hoifung
 * Poon in the Department of Computer Science and
 * Engineering at the University of Washington".
 * 
 * 4. Your publications acknowledge the use or
 * contribution made by the Software to your research
 * using the following citation(s): 
 * Stanley Kok, Parag Singla, Matthew Richardson and
 * Pedro Domingos (2005). "The Alchemy System for
 * Statistical Relational AI", Technical Report,
 * Department of Computer Science and Engineering,
 * University of Washington, Seattle, WA.
 * http://www.cs.washington.edu/ai/alchemy.
 * 
 * 5. Neither the name of the University of Washington nor
 * the names of its contributors may be used to endorse or
 * promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON
 * AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY
 * OF WASHINGTON OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
#include "groundclause.h"
#include "groundpredicate.h"
#include "clause.h"

GroundClause::GroundClause(const Clause* const & c) 
  : numSatLiterals_(NULL), wt_(c->getWt()), parentWtPtrs_(NULL), foClauseFrequencies_(NULL)
{
  int numPreds = c->getNumPredicates();
  intArrRep_ = new Array<unsigned int>;
  gndPreds_ = new Array<GroundPredicate*>(numPreds);
  gndPreds_->growToSize(numPreds);
  gndPredSenses_ = new Array<bool>(numPreds);
  gndPredSenses_->growToSize(numPreds);
  gndPredIndexes_ = new Array<int>(numPreds);
  gndPredIndexes_->growToSize(numPreds);
  
    // for each predicate in clause c
  for (int i = 0; i < numPreds; i++)
  {
      // add the predicate's intArrRep to GroundClause's intArrRep_
    Predicate* pred = c->getPredicate(i);
    bool predSense = pred->getSense();
    GroundPredicate* gndPred = new GroundPredicate(pred);
    if (predSense) intArrRep_->append(1);
    else           intArrRep_->append((unsigned int) 0);
    gndPred->appendIntArrRep(*intArrRep_);
    (*gndPreds_)[i] = gndPred;
    (*gndPredSenses_)[i] = predSense;
    (*gndPredIndexes_)[i] = -1; //set this later
  }

  unsigned int hashCode = Hash::hash(*intArrRep_);
  intArrRep_->append(hashCode);
  intArrRep_->compress();

  for (int i = 0; i < numPreds; i++)
  {
    bool ok = (*gndPreds_)[i]->appendGndClause(this, (*gndPredSenses_)[i]);
    assert(ok); ok = true;  // avoid compilation warning
  }
}


void GroundClause::deleteAllGndPreds()
{
  for (int i = 0; i < gndPreds_->size(); i++)
    delete (*gndPreds_)[i];
}


void GroundClause::replaceGroundPredicate(const int& i, 
                                          GroundPredicate* const & gndPred)
{
  GroundPredicate* old = (*gndPreds_)[i];
  (*gndPreds_)[i] = gndPred;
  delete old;
}


  //the GroundPredicates that make up this GroundClause must already have 
  //their truth values set
void GroundClause::initNumSatLiterals(const int& numChains)
{ 
  if(numSatLiterals_) delete numSatLiterals_;
  numSatLiterals_ = new Array<int>;
  numSatLiterals_->growToSize(numChains, 0);
  assert(gndPreds_->size() == gndPredSenses_->size());
  for (int c = 0; c < numChains; c++)
    for (int i = 0; i < gndPreds_->size(); i++)
    {
      if( (*gndPreds_)[i]->getTruthValue(c) == (*gndPredSenses_)[i] )
      {
        (*numSatLiterals_)[c]++;
        assert((*numSatLiterals_)[c] <= gndPreds_->size());
      }
      
    }
}


void GroundClause::printWithoutWt(ostream& out, 
                                  const Domain* const & domain) const
{
  for (int i = 0; i < gndPreds_->size(); i++)
  {
    if (!(*gndPredSenses_)[i]) out << "!";
    (*gndPreds_)[i]->print(out,domain);
    if (i < gndPreds_->size()-1) out << " v ";
  }
}


void GroundClause::printWithoutWt(ostream& out) const
{
  for (int i = 0; i < gndPreds_->size(); i++)
  {
    if (!(*gndPredSenses_)[i]) out << "!";
    (*gndPreds_)[i]->print(out);
    if (i < gndPreds_->size()-1) out << " v ";
  }
}


void GroundClause::print(ostream& out, const Domain* const & domain) const
{ out << wt_ << " "; printWithoutWt(out,domain); }


void GroundClause::print(ostream& out) const
{ out << wt_ << " "; printWithoutWt(out); }


/**
 * Computes and returns the size of this ground clause.
 */
double GroundClause::sizeKB()
{
  double size = 0;
    //intArrRep_
  if (intArrRep_)
    size += (intArrRep_->size()*sizeof(unsigned int) / 1024.0);
    // gndPreds_
  //if (gndPreds_)
  //{
  //  for (int i = 0; i < gndPreds_->size(); i++)
  //    size += (((*gndPreds_)[i])->sizeKB());
  //}
    // gndPredSenses_
  if (gndPredSenses_)
    size += (gndPredSenses_->size()*sizeof(bool) / 1024.0);
    // gndPredIndexes_
  if (gndPredIndexes_)
    size += (gndPredIndexes_->size()*sizeof(int) / 1024.0);
    // numSatLiterals_
  if (numSatLiterals_)
    size += (numSatLiterals_->size()*sizeof(int) / 1024.0);
    // wt_
  size += (sizeof(double) / 1024.0);
    // threshold_
  size += (sizeof(double) / 1024.0);
    // parentWtPtrs_
  if (parentWtPtrs_)
    size += (parentWtPtrs_->size()*sizeof(const double*) / 1024.0);
    // foClauseFrequencies_
  if (foClauseFrequencies_)
    size += (foClauseFrequencies_->size()*2*sizeof(int) / 1024.0);
  
  return size;
}

