/*
 * All of the documentation and software included in the
 * Alchemy Software is copyrighted by Stanley Kok, Parag
 * Singla, Matthew Richardson, Pedro Domingos, Marc
 * Sumner and Hoifung Poon.
 * 
 * Copyright [2004-06] Stanley Kok, Parag Singla, Matthew
 * Richardson, Pedro Domingos, Marc Sumner and Hoifung
 * Poon. All rights reserved.
 * 
 * Contact: Pedro Domingos, University of Washington
 * (pedrod@cs.washington.edu).
 * 
 * Redistribution and use in source and binary forms, with
 * or without modification, are permitted provided that
 * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above
 * copyright notice, this list of conditions and the
 * following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the
 * above copyright notice, this list of conditions and the
 * following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use
 * of this software must display the following
 * acknowledgment: "This product includes software
 * developed by Stanley Kok, Parag Singla, Matthew
 * Richardson, Pedro Domingos, Marc Sumner and Hoifung
 * Poon in the Department of Computer Science and
 * Engineering at the University of Washington".
 * 
 * 4. Your publications acknowledge the use or
 * contribution made by the Software to your research
 * using the following citation(s): 
 * Stanley Kok, Parag Singla, Matthew Richardson and
 * Pedro Domingos (2005). "The Alchemy System for
 * Statistical Relational AI", Technical Report,
 * Department of Computer Science and Engineering,
 * University of Washington, Seattle, WA.
 * http://www.cs.washington.edu/ai/alchemy.
 * 
 * 5. Neither the name of the University of Washington nor
 * the names of its contributors may be used to endorse or
 * promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON
 * AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY
 * OF WASHINGTON OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
#include <unistd.h>
#include <fstream>
#include <climits>
#include <sys/times.h>
#include "fol.h"
#include "arguments.h"
#include "util.h"
#include "learnwts.h"
#include "infer.h"
#include "maxwalksat.h"
#include "mcsat.h"

extern const char* ZZ_TMP_FILE_POSTFIX; //defined in fol.y

char* ainMLNFiles     = NULL;
char* aevidenceFiles  = NULL;
char* aresultsFile    = NULL;
char* aqueryPredsStr  = NULL;
char* aqueryFile      = NULL;
//bool  aclosedWorld   = false; 
char* aclosedWorldPredsStr = NULL;
char* aopenWorldPredsStr = NULL;
bool  amapPos = false;
bool  amapAll = false;
bool  agibbsInfer = false;
int   asmooth = 1;

bool  amcsat = false;	// mc-sat
bool  asimtp = false;	// simulated tempering
int hardWt = -1;		// soften hard clauses with the given wt if specified

int  mwsMaxSteps = 100000;
int  mwsTries    = 1;
int  mwsTargetWt = 0;
bool mwsHard     = false;
int mwsHeuristic = 1; // 0 = RANDOM, 1 = BEST, 2 = TABU
int mwsTabuLength = 5;
bool lazy = false;
//int mwsSeed = -1;
int seed = -1;
// Limit in kbytes before using lazy version
int mwsLimit = -1;
bool lazyNoApprox = false;

// MC-SAT / SampleSat params
int numStepsEveryMCSat = 1;	// number of total steps (mcsat & Gibbs) for each MC-SAT step
int numsol = 10;		// the nth solution to return
int ws_noise = 15;		// ws noise
int ws_restart = 10;	// ws restart = numtry
int ws_cutoff = 100000;	// ws cutoff
int saRatio = 50;		// percent of SA steps
int saTemp = 50;	    // temperature/100: SA temperature
bool lateSa = false;	// sa only at a plateur

// Simulated Tempering params
int subInterval = 2;
int numST = 3;
int numSwap = 10;

int    mcmcNumChains    = 10;
int    mcmcBurnMinSteps = 100;
int    mcmcBurnMaxSteps = 100;
double mcmcDelta        = 0.05;
double mcmcEpsilonError = 0.01;
int    mcmcMinSteps     = -1;
int    mcmcMaxSteps     = 1000;
double mcmcFracConverged= 0.95;
int    mcmcWalksatType  = 1;
int    mcmcMaxSeconds   = -1;
int    mcmcSamplesPerTest= 100;


ARGS ARGS::Args[] = 
{
  ARGS("i", ARGS::Req, ainMLNFiles, "Comma-separated input .mln files."),

  ARGS("e", ARGS::Req, aevidenceFiles, 
       "Comma-separated .db files containing known ground atoms (evidence), "
       "including function definitions."),

  ARGS("r", ARGS::Req, aresultsFile,
       "The probability estimates are written to this file."),
    
  ARGS("q", ARGS::Opt, aqueryPredsStr, 
       "Query atoms (comma-separated with no space)  "
       ",e.g., cancer,smokes(x),friends(Stan,x). Query atoms are always "
       "open world."),

  ARGS("f", ARGS::Opt, aqueryFile,
       "A .db file containing ground query atoms, "
       "which are are always open world."),

  //ARGS("c", ARGS::Tog, aclosedWorld,
  //     "If specified, all atoms are closed-world except the query ones; "
  //     "otherwise, all atoms are open world. "
  //     "Cannot be used with the -o option."),

  ARGS("c", ARGS::Opt, aclosedWorldPredsStr,
       "Specified non-evidence atoms (comma-separated with no space) are closed world, "
       "otherwise, all non-evidence atoms are open world. "
       "Atoms appearing here cannot be query atoms and cannot appear in the -o option."),

  ARGS("o", ARGS::Opt, aopenWorldPredsStr,
       "Specified evidence atoms (comma-separated with no space) are open world, "
       "while other evidence atoms are closed-world. "
       "Atoms appearing here cannot appear in the -c option."),

  ARGS("m", ARGS::Tog, amapPos, 
       "Run MAP inference and return only positive query atoms."),

  ARGS("a", ARGS::Tog, amapAll, 
       "Run MAP inference and show 0/1 results for all query atoms."),

  ARGS("p", ARGS::Tog, agibbsInfer, 
       "Run inference using MCMC (Gibbs sampling) and return probabilities "
       "for all query atoms."),
  
  ARGS("ms", ARGS::Tog, amcsat,
       "Run inference using MC-SAT and return probabilities "
       "for all query atoms"),

  ARGS("simtp", ARGS::Tog, asimtp,
       "Run inference using simulated tempering and return probabilities "
       "for all query atoms"),

  ARGS("seed", ARGS::Opt, seed,
       "[random] Seed used to initialize the randomizer in the inference algorithm. "
       "If not set, seed is initialized from the current date and time."),

  ARGS("mwsMaxSteps", ARGS::Opt, mwsMaxSteps,
       "[1000000] The max number of steps taken by MaxWalkSat."),

  ARGS("mwsTries", ARGS::Opt, mwsTries, 
       "[1] The max number of attempts MaxWalkSat takes to find a solution."),

  ARGS("mwsTargetWt", ARGS::Opt, mwsTargetWt,
       "[the best possible] MaxWalkSat tries to find a solution with weight <= "
       "specified weight."),

  ARGS("mwsHard", ARGS::Opt, mwsHard, 
       "[false] MaxWalkSat never breaks a hard clause in order to satisfy "
       "a soft one."),
  
  ARGS("mwsLimit", ARGS::Opt, mwsLimit, 
       "[-1] Maximum limit in kbytes which MaxWalkSat should use for inference. "
	   "-1 means main memory available on system is used."),

  ARGS("mwsHeuristic", ARGS::Opt, mwsHeuristic,
       "[1] Heuristic used in MaxWalkSat (0 = RANDOM, 1 = BEST, 2 = TABU, 3 = SAMPLESAT)."),
  
  ARGS("mwsTabuLength", ARGS::Opt, mwsTabuLength,
       "[5] Minimum number of flips between flipping the same atom when using "
       "the tabu heuristic in MaxWalkSat." ),

  ARGS("lazy", ARGS::Opt, lazy, 
       "[false] Run lazy version of MAP inference / MC-SAT if this flag is set."),
  
  ARGS("lazyNoApprox", ARGS::Opt, lazyNoApprox, 
       "[false] LazySat will not approximate by deactivating atoms to save memory. This flag "
	   "is ignored if -lazy is not set."),
  
  ARGS("mcmcSmooth", ARGS::Opt, asmooth, 
       "[1] Number of samples to use for smoothing."),

  ARGS("simtpSubInterval", ARGS::Opt, subInterval,
		"[2] Selection interval btw swap attempts"),

  ARGS("simtpNumRuns", ARGS::Opt, numST,
		"[3] Number of sim. tempering runs"),

  ARGS("simtpNumSwap", ARGS::Opt, numSwap,
		"[10] Number of swapping chains"),

  ARGS("numStepsEveryMCSat", ARGS::Opt, numStepsEveryMCSat,
       "[1] Number of total steps (mcsat & gibbs) for every mcsat step"),

  ARGS("mcsatSaRatio", ARGS::Opt, saRatio,
       "[50] Ratio of sim. annealing steps mixed with WalkSAT in MC-SAT"),

  ARGS("mcsatSaTemperature", ARGS::Opt, saTemp,
		"[10] Temperature (/100) for sim. annealing"),

  ARGS("mcsatLateSa", ARGS::Tog, lateSa,
       "[false] Run sim. annealing from the start in samplesat"),

  ARGS("mcsatNumSolutions", ARGS::Opt, numsol,
       "[10] Return nth SAT solution in samplesat"),

  ARGS("mcmcNumChains", ARGS::Opt, mcmcNumChains, 
       "[10] Number of MCMC chains (there must be at least 2)."),

  ARGS("mcmcBurnMinSteps", ARGS::Opt, mcmcBurnMinSteps,
       "[100] Minimun number of burn in steps (-1: no minimum)."),

  ARGS("mcmcBurnMaxSteps", ARGS::Opt, mcmcBurnMaxSteps,
       "[100] Maximum number of burn-in steps (-1: no maximum)."),

  ARGS("mcmcDelta", ARGS::Opt, mcmcDelta,
       "[0.05] During Gibbs sampling, probabilty that epsilon error is "
       "exceeded is less than this value."),

  ARGS("mcmcEpsilonError", ARGS::Opt, mcmcEpsilonError,
       "[0.01] Fractional error from true probability."),

  ARGS("mcmcMinSteps", ARGS::Opt, mcmcMinSteps, 
       "[-1] Minimum number of Gibbs sampling steps."),

  ARGS("mcmcMaxSteps", ARGS::Opt, mcmcMaxSteps, 
       "[1000] Maximum number of Gibbs sampling steps."),

  ARGS("mcmcFracConverged", ARGS::Opt, mcmcFracConverged, 
       "[0.95] Fraction of ground atoms with probabilities that "
       "have converged."),

  ARGS("mcmcWalksatType", ARGS::Opt, mcmcWalksatType, 
       "[1] Use Max Walksat to initialize ground atoms' truth values "
       "(1: use Max Walksat, 0: random initialization)."),

  ARGS("mcmcMaxSeconds", ARGS::Opt, mcmcMaxSeconds, 
       "[-1] Max number of seconds to run MCMC (-1: no maximum)."),
  
  ARGS("mcmcSamplesPerTest", ARGS::Opt, mcmcSamplesPerTest, 
       "[100] Perform convergence test once after this many number of samples "
       "per chain."),

  ARGS()
};


int main(int argc, char* argv[])
{
  ///////////////////////////  read & prepare parameters ///////////////////////
  ARGS::parse(argc, argv, &cout);

  Timer timer;
  double begSec = timer.time(); 

  string inMLNFile, wkMLNFile, evidenceFile, queryPredsStr, queryFile;

  StringHashArray queryPredNames;
  StringHashArray owPredNames;
  StringHashArray cwPredNames;
  Domain* domain = NULL;
  MLN* mln = NULL;
  GroundPredicateHashArray queries;
  GroundPredicateHashArray knownQueries;
  Array<string> constFilesArr;
  Array<string> evidenceFilesArr;

  Array<Predicate *> queryPreds;
  Array<TruthValue> queryPredValues;
  
  //the second .mln file to the last one in ainMLNFiles _may_ be used 
  //to hold constants, so they are held in constFilesArr. They will be
  //included into the first .mln file.

    //extract .mln, .db file names
  extractFileNames(ainMLNFiles, constFilesArr);
  assert(constFilesArr.size() >= 1);
  inMLNFile.append(constFilesArr[0]);
  constFilesArr.removeItem(0);
  extractFileNames(aevidenceFiles, evidenceFilesArr);
  
  if (aqueryPredsStr) queryPredsStr.append(aqueryPredsStr);
  if (aqueryFile) queryFile.append(aqueryFile);

  if (queryPredsStr.length() == 0 && queryFile.length() == 0)
  { cout << "No query predicates specified" << endl; return -1; }

  ofstream resultsOut(aresultsFile);
  if (!resultsOut.good())
  { cout << "ERROR: unable to open " << aresultsFile << endl; return -1; }

  if (mcmcNumChains < 2) 
  { cout << "ERROR: there must be at least 2 MCMC chains" << endl; return -1; }

  if (!asimtp && !amapPos && !amapAll && !agibbsInfer && !amcsat)
  { cout << "ERROR: must specify one of -ms/-simtp/-m/-a/-p flags." << endl; return-1; }

  if (lazy && !(amapPos || amapAll || amcsat))
  {
  	cout << "ERROR: -lazy can only be used with MAP inference "
  	  		"or MC-SAT (-m, -a, or -ms option)." << endl;
  	return-1;
  }

  //if (aclosedWorld && aopenWorldPredsStr)
  //{ cout << "ERROR: cannot specify both -c & -o together." << endl; return -1; }

    //extract names of all query predicates
  if (queryPredsStr.length() > 0 || queryFile.length() > 0)
  {
    if (!extractPredNames(queryPredsStr, &queryFile, queryPredNames)) return -1;
  }

  if (mwsMaxSteps <= 0)
  { cout << "ERROR: mwsMaxSteps must be positive" << endl; return -1; }

  if (mwsTries <= 0)
  { cout << "ERROR: mwsTries must be positive" << endl; return -1; }

    //extract names of open-world evidence predicates
  if (aopenWorldPredsStr)
  {
    if (!extractPredNames(string(aopenWorldPredsStr), NULL, owPredNames)) 
      return -1;
    assert(owPredNames.size() > 0);
  }

    //extract names of closed-world non-evidence predicates
  if (aclosedWorldPredsStr)
  {
    if (!extractPredNames(string(aclosedWorldPredsStr), NULL, cwPredNames)) 
      return -1;
    assert(cwPredNames.size() > 0);
    if (!checkQueryPredsNotInClosedWorldPreds(queryPredNames, cwPredNames))
    return -1;
  }

  // TODO: Check if atom in -c and -o -> error


  // TODO: Check if ev. atom in -c or
  // non-evidence in -o -> warning (this is default)



  SampleSatParams* ssparams = new SampleSatParams;
  ssparams->lateSa = lateSa;
  ssparams->saRatio = saRatio;
  ssparams->saTemp = saTemp;

  MaxWalksatParams* mwsparams = NULL;
  mwsparams = new MaxWalksatParams;
  mwsparams->ssParams = ssparams;
  mwsparams->maxSteps = mwsMaxSteps;
  mwsparams->maxTries = mwsTries;
  mwsparams->targetCost = mwsTargetWt;
  mwsparams->hard = mwsHard;
    // numSolutions only applies when used in SampleSat.
    // When just MWS, this is set to 1
  mwsparams->numSolutions = numsol;
  mwsparams->heuristic = mwsHeuristic;
  mwsparams->tabuLength = mwsTabuLength;

  MCSatParams* msparams = new MCSatParams;
  msparams->mwsParams = mwsparams;
    // MC-SAT needs only one chain
  msparams->numChains = 1;
  msparams->burnMinSteps = mcmcBurnMinSteps;
  msparams->burnMaxSteps = mcmcBurnMaxSteps;
  msparams->minSteps = mcmcMinSteps;
  msparams->maxSteps = mcmcMaxSteps;
  
  //////////////////// read in clauses & evidence predicates //////////////////

  cout << "Reading formulas and evidence predicates..." << endl;

    //copy inMLNFile to workingMLNFile & app '#include "evid.db"'
  string::size_type bslash = inMLNFile.rfind("/");
  string tmp = (bslash == string::npos) ? 
               inMLNFile:inMLNFile.substr(bslash+1,inMLNFile.length()-bslash-1);
  char buf[100];
  sprintf(buf, "%s%s", tmp.c_str(), ZZ_TMP_FILE_POSTFIX);
  wkMLNFile = buf;
  copyFileAndAppendDbFile(inMLNFile, wkMLNFile,
                          evidenceFilesArr, constFilesArr);

    // parse wkMLNFile, and create the domain, MLN, database
  domain = new Domain;
  mln = new MLN();
  bool addUnitClauses = false;
  bool warnAboutDupGndPreds = true;
  bool mustHaveWtOrFullStop = true;
  bool flipWtsOfFlippedClause = true;
  bool allPredsExceptQueriesAreCW = true;
  Domain* forCheckingPlusTypes = NULL;

  if (!allPredsExceptQueriesAreCW && owPredNames.empty())
  {
    Array<string> pnames;
    domain->getNonEqualPredicateNames(pnames);
    for (int i = 0; i < pnames.size(); i++) owPredNames.append(pnames[i]);
  }

	// Parse as if lazy inference is set to true to set evidence atoms in DB
    // If lazy is not used, this is removed from DB
  if (!runYYParser(mln, domain, wkMLNFile.c_str(), allPredsExceptQueriesAreCW, 
                   &owPredNames, &queryPredNames, addUnitClauses, 
                   warnAboutDupGndPreds, 0, mustHaveWtOrFullStop, 
                   forCheckingPlusTypes, true, flipWtsOfFlippedClause))
  {
    unlink(wkMLNFile.c_str());
    return -1;
  }

  unlink(wkMLNFile.c_str());

    //////////////////////////// run inference /////////////////////////////////

    ///////////////////////// read & create query predicates ///////////////////
  if (queryFile.length() > 0)
  {
    cout <<"Reading query predicates that are specified in query file..."<<endl;
    bool ok = createQueryFilePreds(queryFile, domain, domain->getDB(), &queries,
                                   &knownQueries);
    if (!ok) { cout <<"Failed to create query predicates." << endl; exit(-1); }
  }

  Array<int> allPredGndingsAreQueries;
  allPredGndingsAreQueries.growToSize(domain->getNumPredicates(), false);
  if (queryPredsStr.length() > 0)
  {
    cout << "Creating query predicates that are specified on command line..." 
         << endl;
    bool ok = createComLineQueryPreds(queryPredsStr, domain, domain->getDB(), 
                                      &queries, &knownQueries, 
                                      &allPredGndingsAreQueries);
    if (!ok) { cout <<"Failed to create query predicates." << endl; exit(-1); }
  }

    // Create inference algorithm and state based on queries and mln / domain
  VariableState* state = new VariableState(&queries, &allPredGndingsAreQueries,
                                           mln, domain, lazy);
  Inference* inference = NULL;

    // MAP inference or MC-SAT (types of inference for which a
    // lazy version exists)
  if (amapPos || amapAll || amcsat) 
  {
    if (amapPos || amapAll)
    {
        // When standalone MWS, numSolutions is always 1
        // (maybe there is a better way to do this?)
      mwsparams->numSolutions = 1;
      inference = new MaxWalkSat(state, seed, mwsparams);
    }
    else if (amcsat)
    {
      inference = new MCSAT(state, seed, msparams);
    }
      
    inference->init();
    inference->infer();
    
    if (amapPos)
      state->printTruePreds(resultsOut);
    else
        // Print out query probs to results file
      state->printProbabilities(resultsOut);
  }

  resultsOut.close();
  if (mwsparams) delete mwsparams;
  if (ssparams) delete ssparams;
  delete domain;
  for (int i = 0; i < knownQueries.size(); i++)  delete knownQueries[i];
  delete inference;
  delete state;
  
  cout << "total time taken = "; Timer::printTime(cout, timer.time()-begSec);
  cout << endl;
}
