/*
 * All of the documentation and software included in the
 * Alchemy Software is copyrighted by Stanley Kok, Parag
 * Singla, Matthew Richardson, Pedro Domingos, Marc
 * Sumner and Hoifung Poon.
 * 
 * Copyright [2004-06] Stanley Kok, Parag Singla, Matthew
 * Richardson, Pedro Domingos, Marc Sumner and Hoifung
 * Poon. All rights reserved.
 * 
 * Contact: Pedro Domingos, University of Washington
 * (pedrod@cs.washington.edu).
 * 
 * Redistribution and use in source and binary forms, with
 * or without modification, are permitted provided that
 * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above
 * copyright notice, this list of conditions and the
 * following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the
 * above copyright notice, this list of conditions and the
 * following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use
 * of this software must display the following
 * acknowledgment: "This product includes software
 * developed by Stanley Kok, Parag Singla, Matthew
 * Richardson, Pedro Domingos, Marc Sumner and Hoifung
 * Poon in the Department of Computer Science and
 * Engineering at the University of Washington".
 * 
 * 4. Your publications acknowledge the use or
 * contribution made by the Software to your research
 * using the following citation(s): 
 * Stanley Kok, Parag Singla, Matthew Richardson and
 * Pedro Domingos (2005). "The Alchemy System for
 * Statistical Relational AI", Technical Report,
 * Department of Computer Science and Engineering,
 * University of Washington, Seattle, WA.
 * http://www.cs.washington.edu/ai/alchemy.
 * 
 * 5. Neither the name of the University of Washington nor
 * the names of its contributors may be used to endorse or
 * promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON
 * AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY
 * OF WASHINGTON OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
#ifndef VARIABLESTATE_H_
#define VARIABLESTATE_H_

#include "mrf.h"

#define NOVALUE -1

/**
 * Represents the state of propositional variables and clauses. Some of this
 * code is based on the MaxWalkSat package of Kautz et al.
 * 
 * All inference algorithms should have a VariableState to access the
 * information needed in its predicates and clauses.
 * 
 * Each atom has its own index starting at 1. The negation of an atom with
 * index a is represented by -a (this is why the indices do not start at 0).
 * Each clause has its own index starting at 0.
 * 
 * A VariableState is either eager or lazy. Eager states build an MRF upfront
 * based on an MLN and a domain. Thus, all ground clauses and predicates are
 * in memory after building the MRF. Lazy states activate atoms and clauses as
 * they are needed from the MLN and domain. An atom is activated if it is in
 * an unsatisfied clause with the assumption of all atoms being false or if it
 * is looked at during inference (it is flipped or the cost of flipping it is
 * computed). Active clauses are those which contain active atoms.
 */
class VariableState
{
 public:

  /**
   * Constructor for a VariableState. The hard clause weight is set. In lazy
   * mode, the initial active atoms and clauses are retrieved and in eager
   * mode, the MRF is built and the atoms and clauses are retrieved from it.
   * In addition, all information arrays are filled with information.
   * 
   * @param queries Query predicates used to build MRF in eager mode.
   * @param allPredGndingsAreQueries Array used to build MRF in eager mode.
   * @param mln mln and domain are used to build MRF in eager state and to
   * retrieve active atoms in lazy state.
   * @param domain mln and domain are used to build MRF in eager state and to
   * retrieve active atoms in lazy state.
   * @param lazy Flag stating whether lazy mode is used or not.
   */
  VariableState(const GroundPredicateHashArray* const& queries,
                const Array<int>* const & allPredGndingsAreQueries,
                const MLN* const & mln, const Domain* const & domain,
                const bool& lazy)
  {
    this->mln_ = (MLN*)mln;
    this->domain_ = (Domain*)domain;
    this->lazy = lazy;

      // Instantiate information
    numAtoms = 0;
    baseNumAtoms = 0;
    activeAtoms = 0;
    numClauses = 0;
    numFalseClauses = 0;
    costOfFalseClauses = 0.0;
    lowCost = DBL_MAX;
    lowBad = INT_MAX;

      // Clauses and preds are stored in gndClauses_ and gndPreds_
    gndClauses_ = new Array<GroundClause*>;
    gndPreds_ = new Array<GroundPredicate*>;

      // Set the hard clause weight
    setHardClauseWeight();

      // Lazy version: Produce state with initial active atoms and clauses
    if (lazy)
    {
        // Unknown preds are treated as false
      domain_->getDB()->setPerformingInference(true);

        // Blocks are copied from the domain and predicates are added
        // to the state as active
      initLazyBlocks();

      clauseLimit = INT_MAX;
      noApprox = false;
      haveDeactivated = false;

        ///// Get initial active atoms and clauses /////

        // Get initial set of active atoms (atoms in unsat. clauses)
        // TODO: Ignore active preds because of block variables already activated?
      getActiveClauses(newClauses_);
      int defaultCnt = newClauses_.size();
      long double defaultCost = 0;

      for (int i = 0; i < defaultCnt; i++)
      {
        if (newClauses_[i]->isHardClause())
          defaultCost += hardWt;
        else
          defaultCost += newClauses_[i]->getWt();
        //newClauses_[i]->deleteAllGndPreds();
        delete newClauses_[i];
      }
      newClauses_.clear();

      baseNumAtoms = gndPredHashArray_.size();
      cout << "Number of Baseatoms = " << baseNumAtoms << endl;
      cout<<"Default => Cost\t"<<"******\t"<<" Clause Cnt\t"<<endl;
      cout<<"           "<<defaultCost<<"\t"<<"******\t"<<defaultCnt<<"\t"<<endl<<endl;

        // Set base atoms as active in DB
      for (int i = 0; i < baseNumAtoms; i++)
      {
        domain_->getDB()->setActiveStatus(gndPredHashArray_[i], true);
        activeAtoms++;        
      }

        // Get the initial set of active clauses
      getActiveClauses(newClauses_);
    } // End lazy version
      // Eager version: Use KBMC to produce the state
    else
    {
        // TODO: Is this different for different inference algorithms?
      bool markHardGndClauses = true;
      bool trackParentClauseWts = false;
      mrf_ = new MRF(queries, allPredGndingsAreQueries, domain_, domain_->getDB(),
                     mln_, markHardGndClauses, trackParentClauseWts, -1);

        //delete to save space. Can be deleted because no more hashing of
        //gndClauses are required beyond this point
      mrf_->deleteGndClausesIntArrReps();
        //delete to save space. Can be deleted because no more gndClauses are
        //appended to gndPreds beyond this point
      mrf_->deleteGndPredsGndClauseSets();
        //do not delete the intArrRep in gndPreds_;
      //domain->deleteDB();
      //delete mln;

        // Blocks built in MRF
      blocks_ = mrf_->getBlocks();
      blockEvidence_ = mrf_->getBlockEvidence();
        
        // Put ground clauses in newClauses_
      newClauses_ = *(Array<GroundClause*>*)mrf_->getGndClauses();
        // Put ground preds in the hash array
      const Array<GroundPredicate*>* gndPreds = mrf_->getGndPreds();
      for (int i = 0; i < gndPreds->size(); i++)
        gndPredHashArray_.append((*gndPreds)[i]);
    
        // baseNumAtoms are all atoms in eager version
      baseNumAtoms = gndPredHashArray_.size();        
    } // End eager version
    
      // At this point, ground clauses are held in newClauses_
      // and ground predicates are held in gndPredHashArray_
      // for both versions
    
      // Add the clauses and preds and fill info arrays
    bool initial = true;
    addNewClauses(initial);
  }

  /**
   * Destructor.
   */ 
  ~VariableState()
  {
    if (lazy)
    {
        // Block information from lazy version is deleted
      for (int i = 0; i < blocks_->size(); i++)
        (*blocks_)[i].clearAndCompress();
      delete blocks_;
    
      delete blockEvidence_;  
    }
    else
    {
        // MRF from eager version is deleted
      if (mrf_) delete mrf_;
    } 
  }


  /**
   * New clauses are added to the state. If not the initialization, then
   * makecost and breakcost are updated for the new atoms.
   * 
   * @param initial If true, this is the first time new clauses
   * have been added.
   */
  void addNewClauses(bool initial)
  {
    //cout<<"adding new clauses.."<<endl;

      // Store the old number of clauses and atoms
    int oldNumClauses = numClauses;
    int oldNumAtoms = numAtoms;

    gndClauses_->append(newClauses_);
    gndPreds_->growToSize(gndPredHashArray_.size());
    numAtoms = gndPreds_->size();
    numClauses = gndClauses_->size();
    
      // If no new atoms or clauses have been added, then do nothing
    if (numAtoms == oldNumAtoms && numClauses == oldNumClauses) return;

      // atomIdx starts at 1
    atom.growToSize(numAtoms + 1, false);
    makeCost.growToSize(numAtoms + 1, 0.0);
    breakCost.growToSize(numAtoms + 1, 0.0);
    lowAtom.growToSize(numAtoms + 1, false);
    fixedAtom.growToSize(numAtoms + 1, 0);

      // Copy ground preds to gndPreds_ and set values in atom and lowAtom
    for (int i = oldNumAtoms; i < gndPredHashArray_.size(); i++)
    {
      (*gndPreds_)[i] = gndPredHashArray_[i];
      if (lazy)
      {
        lowAtom[i + 1] = atom[i + 1] = 
          (domain_->getDB()->getValue((*gndPreds_)[i]) == TRUE) ? true : false;
      }
    }
    newClauses_.clear();

    clause.growToSize(numClauses);
    clauseCost.growToSize(numClauses);
    clauseSize.growToSize(numClauses);
    falseClause.growToSize(numClauses);
    whereFalse.growToSize(numClauses);
    numTrueLits.growToSize(numClauses);
    watch1.growToSize(numClauses);
    watch2.growToSize(numClauses);
    isSatisfied.growToSize(numClauses, false);

    occurence.growToSize(2*numAtoms + 1);

    for (int i = oldNumClauses; i < numClauses; i++)
    {
      GroundClause* gndClause = (*gndClauses_)[i];
      int numGndPreds = gndClause->getNumGroundPredicates();
      clauseSize[i] = numGndPreds;
      clause[i].growToSize(numGndPreds);

      for (int j = 0; j < numGndPreds; j++) 
      {
          // idx in gndClause + 1 (negated if neg. literal)
        int idx = gndClause->getGroundPredicateIndex(j);
        assert(idx >= 0);
        int lit = (gndClause->getGroundPredicateSense(j)) ?
                  idx + 1 : -(idx + 1);
        clause[i][j] = lit;
          //cout << "numAtoms " << numAtoms << endl;
        int litIdx = 2*abs(lit) - (lit > 0);
        occurence[litIdx].append(i);
      }

        // Hard clause weights has been previously determined
      if (gndClause->isHardClause())
        clauseCost[i] = hardWt;
      else
        clauseCost[i] = gndClause->getWt();
    }

    if (!initial)
      initMakeBreakCostWatch(oldNumClauses);
    //cout<<"done adding new clauses.."<<endl;
  }

  
  /**
   * Makes a random truth assigment to all (active) atoms. Blocks are
   * taken into account: exactly one atom in the block is set to true
   * and the others are set to false.
   */
  void initRandom()
  {
    for (int i = 0; i < numClauses; i++) numTrueLits[i] = 0;
    
      // For each block: select one to set to true
    for (int i = 0; i < blocks_->size(); i++)
    {
      Array<int>& block = (*blocks_)[i];
      
        // True fixed atom in the block: do nothing
      if (trueFixedAtomInBlock(i))
        continue;

        // If evidence atom exists, then all others are false
      if ((*blockEvidence_)[i])
      {
          // If first argument is -1, then all are set to false
        setOthersInBlockToFalse(-1, i);
        continue;
      }
      
      int chosen = random() % block.size();
      setValueOfAtom(block[chosen] + 1, true);
      setOthersInBlockToFalse(chosen, i);
    }
      
      // Random truth value for all not in blocks
    for (int i = 1; i <= baseNumAtoms; i++)
    {
        // fixedAtom[i] = -1: false, fixedAtom[i] = 1: true
      if (fixedAtom[i] != 0) setValueOfAtom(i, (fixedAtom[i] == 1));
        // Blocks are initialized above
      if (getBlockIdx(i - 1) >= 0) continue;
        // Not fixed and not in block
      else setValueOfAtom(i, random() % 2);
      //cout << atom[i] << " ";
    }
      // Initialize info arrays
    initMakeBreakCostWatch(0);
  }

  /**
   * Initialize makeCost and breakCost and watch arrays based on the current
   * variable assignment.
   */
  void initMakeBreakCostWatch(const int& startClause)
  {
    int theTrueLit = -1;
      // Initialize breakCost and makeCost in the following:
    for (int i = startClause; i < numClauses; i++)
    {
      int trueLit1 = 0;
      int trueLit2 = 0;
      long double cost = clauseCost[i];
      numTrueLits[i] = 0;
      for (int j = 0; j < clauseSize[i]; j++)
      {
        if (isTrueLiteral(clause[i][j]))
        { // ij is true lit
          numTrueLits[i]++;
          theTrueLit = abs(clause[i][j]);
          if (!trueLit1) trueLit1 = theTrueLit;
          else if (trueLit1 && !trueLit2) trueLit2 = theTrueLit;
        }
      }
        // Unsatisfied positive-weighted clauses or
        // Satisfied negative-weighted clauses
      if ((numTrueLits[i] == 0 && cost > 0) ||
          (numTrueLits[i] > 0 && cost < 0))
      {
        whereFalse[i] = numFalseClauses;
        falseClause[numFalseClauses] = i;
        numFalseClauses++;
        costOfFalseClauses += abs(cost);
        if (highestCost == abs(cost)) {eqHighest = true; numHighest++;}

          // Unsat. pos. clause: increase makeCost of all atoms
        if (numTrueLits[i] == 0)
          for (int j = 0; j < clauseSize[i]; j++)
          {
            //cout << "Adding " << cost << " to makeCost of " << abs(clause[i][j]) << endl; 
            makeCost[abs(clause[i][j])] += cost;
          }

          // Sat. neg. clause: increase makeCost if one true literal
        if (numTrueLits[i] == 1)
        {
            // Subtract neg. cost
          makeCost[theTrueLit] -= cost;
          watch1[i] = theTrueLit;
        }
        else if (numTrueLits[i] > 1)
        {
          watch1[i] = trueLit1;
          watch2[i] = trueLit2;
        }
      }
        // Pos. clauses satisfied by one true literal
      else if (numTrueLits[i] == 1 && cost > 0)
      {
        breakCost[theTrueLit] += cost;
        watch1[i] = theTrueLit;
      }
        // Pos. clauses satisfied by 2 or more true literals
      else if (cost > 0)
      { /*if (numtruelit[i] == 2)*/
        watch1[i] = trueLit1;
        watch2[i] = trueLit2;
      }
        // Unsat. neg. clauses: increase breakCost of all atoms
      else if (numTrueLits[i] == 0 && cost < 0)
      {
        for (int j = 0; j < clauseSize[i]; j++)
          breakCost[abs(clause[i][j])] -= cost;
      }
    } // for all clauses
  }

  int getNumAtoms() { return numAtoms; }
  
  int getNumClauses() { return numClauses; }
  
  /**
   * Returns the absolute index of a random atom.
   */
  int getIndexOfRandomAtom()
  {
    if (numAtoms == 0) return NOVALUE;
    return random()%numAtoms + 1;
  }

  /**
   * Returns the absolute index of a random atom in a random unsatisfied
   * pos. clause or the absolute index of a random true literal in a random
   * satisfied clause.
   */
  int getIndexOfAtomInRandomFalseClause()
  {
    if (numFalseClauses == 0) return NOVALUE;
    int clauseIdx = falseClause[random()%numFalseClauses];
      // Pos. clause: return index of random atom
    if (clauseCost[clauseIdx] > 0)
      return abs(clause[clauseIdx][random()%clauseSize[clauseIdx]]);
      // Neg. clause: find random true lit
    else
      return getRandomTrueLitInClause(clauseIdx);
  }
  
  /**
   * Returns the index of a random unsatisfied pos. clause or a random
   * satisfied neg. clause.
   */
  int getRandomFalseClauseIndex()
  {
    if (numFalseClauses == 0) return NOVALUE;
    return falseClause[random()%numFalseClauses];
  }
  
  /**
   * Returns the cost of the unsatisfied positive and satisfied negative clauses.
   */
  long double getCostOfFalseClauses()
  {
    return costOfFalseClauses;
  }
  
  /**
   * Returns the number of the unsatisfied positive and
   * satisfied negative clauses.
   */
  int getNumFalseClauses()
  {
    return numFalseClauses;
  }

  /**
   * Returns the truth value of an atom.
   * 
   * @atomIdx Index of atom whose truth value is returned.
   * 
   * @return Truth value of atom.
   */
  bool getValueOfAtom(const int& atomIdx)
  {
    return atom[atomIdx];
  }

  /**
   * Sets the truth value of an atom. If in lazy mode, the truth value
   * is propagated to the database.
   * 
   * @param atomIdx Index of atom whose value is to be set.
   * @param value Truth value being set.
   */
  void setValueOfAtom(const int& atomIdx, const bool& value)
  {
      // If atom already has this value, then do nothing
    if (atom[atomIdx] == value) return;
      // Lazy version: propagate assigment to DB
    if (lazy)
    { 
      GroundPredicate* p = gndPredHashArray_[atomIdx - 1];
      if (value)
      {
        domain_->getDB()->setValue(p, TRUE);
      }
      else
      {
        domain_->getDB()->setValue(p, FALSE);
      }
        // If not active, then activate it
      if (!isActive(atomIdx))
      {
        bool ignoreActivePreds = false;
        activateAtom(atomIdx, ignoreActivePreds);
      }
    }
    atom[atomIdx] = value;
  }

  /**
   * Flip the truth value of an atom and update info arrays.
   * 
   * @param toFlip Index of atom to be flipped.
   */
  void flipAtom(const int& toFlip)
  {
    bool toFlipValue = getValueOfAtom(toFlip);
    register int clauseIdx;
    int sign;
    int oppSign;
    int litIdx;
    if (toFlipValue)
      sign = 1;
    else
      sign = 0;
    oppSign = sign ^ 1;

    flipAtomValue(toFlip);
    
      // Update all clauses in which the atom occurs as a true literal
    litIdx = 2*toFlip - sign;
    Array<int>& posOccArray = getOccurenceArray(litIdx);
    for (int i = 0; i < posOccArray.size(); i++)
    {
      clauseIdx = posOccArray[i];
        // The true lit became a false lit
      int numTrueLits = decrementNumTrueLits(clauseIdx);
      long double cost = getClauseCost(clauseIdx);
      int watch1 = getWatch1(clauseIdx);
      int watch2 = getWatch2(clauseIdx);

        // 1. If last true lit was flipped, then we have to update
        // the makecost / breakcost info accordingly        
      if (numTrueLits == 0)
      {
          // Pos. clause
        if (cost > 0)
        {
            // Add this clause as false in the state
          addFalseClause(clauseIdx);
            // Decrease toFlip's breakcost (add neg. cost)
          addBreakCost(toFlip, -cost);
            // Increase makecost of all vars in clause (add pos. cost)
          addMakeCostToAtomsInClause(clauseIdx, cost);
        }
          // Neg. clause
        else
        {
          assert(cost < 0);
            // Remove this clause as false in the state
          removeFalseClause(clauseIdx);
            // Increase breakcost of all vars in clause (add pos. cost)
          addBreakCostToAtomsInClause(clauseIdx, -cost);        
            // Decrease toFlip's makecost (add neg. cost)
          addMakeCost(toFlip, cost);
        }
      }
        // 2. If there is now one true lit left, then move watch2
        // up to watch1 and increase the breakcost / makecost of watch1
      else if (numTrueLits == 1)
      {
        if (watch1 == toFlip)
        {
          assert(watch1 != watch2);
          setWatch1(clauseIdx, watch2);
          watch1 = getWatch1(clauseIdx);
        }

          // Pos. clause: Increase toFlip's breakcost (add pos. cost)
        if (cost > 0)
        {
          addBreakCost(watch1, cost);
        }
          // Neg. clause: Increase toFlip's makecost (add pos. cost)
        else
        {
          assert(cost < 0);
          addMakeCost(watch1, -cost);
        }
      }
        // 3. If there are 2 or more true lits left, then we have to
        // find a new true lit to watch if one was flipped
      else
      { /* numtruelit[clauseIdx] >= 2 */
          // If watch1[clauseIdx] has been flipped
        if (watch1 == toFlip)
        {
            // find a different true literal to watch
          int diffTrueLit = getTrueLiteralOtherThan(clauseIdx, watch1, watch2);
          setWatch1(clauseIdx, diffTrueLit);
        }
          // If watch2[clauseIdx] has been flipped
        else if (watch2 == toFlip)
        {
            // find a different true literal to watch
          int diffTrueLit = getTrueLiteralOtherThan(clauseIdx, watch1, watch2);
          setWatch2(clauseIdx, diffTrueLit);
        }
      }
    }
        
      // Update all clauses in which the atom occurs as a false literal
    litIdx = 2*toFlip - oppSign;
    Array<int>& negOccArray = getOccurenceArray(litIdx);
    for (int i = 0; i < negOccArray.size(); i++)
    {
      clauseIdx = negOccArray[i];
        // The false lit became a true lit
      int numTrueLits = incrementNumTrueLits(clauseIdx);
      long double cost = getClauseCost(clauseIdx);
      int watch1 = getWatch1(clauseIdx);

        // 1. If this is the only true lit, then we have to update
        // the makecost / breakcost info accordingly        
      if (numTrueLits == 1)
      {
          // Pos. clause
        if (cost > 0)
        {
            // Remove this clause as false in the state
          removeFalseClause(clauseIdx);
            // Increase toFlip's breakcost (add pos. cost)
          addBreakCost(toFlip, cost);        
            // Decrease makecost of all vars in clause (add neg. cost)
          addMakeCostToAtomsInClause(clauseIdx, -cost);
        }
          // Neg. clause
        else
        {
          assert(cost < 0);
            // Add this clause as false in the state
          addFalseClause(clauseIdx);
            // Decrease breakcost of all vars in clause (add neg. cost)
          addBreakCostToAtomsInClause(clauseIdx, cost);
            // Increase toFlip's makecost (add pos. cost)
          addMakeCost(toFlip, -cost);
        }
          // Watch this atom
        setWatch1(clauseIdx, toFlip);
      }
        // 2. If there are now exactly 2 true lits, then watch second atom
        // and update breakcost
      else
      if (numTrueLits == 2)
      {
        if (cost > 0)
        {
            // Pos. clause
            // Decrease breakcost of first atom being watched (add neg. cost)
          addBreakCost(watch1, -cost);
        }
        else
        {
            // Neg. clause
          assert(cost < 0);
            // Decrease makecost of first atom being watched (add neg. cost)
          addMakeCost(watch1, cost);
        }
        
          // Watch second atom
        setWatch2(clauseIdx, toFlip);
      }
    }
  }  

  /**
   * Flips the truth value of an atom. If in lazy mode, the truth value
   * is propagated to the database.
   */  
  void flipAtomValue(const int& atomIdx)
  {
    bool opposite = !atom[atomIdx];
    setValueOfAtom(atomIdx, opposite);
  }

  /**
   * Calculates the improvement achieved by flipping an atom. This is the cost
   * of clauses which flipping the atom makes good minus the cost of clauses which
   * flipping the atom makes bad. In lazy mode, if the atom is not active, then
   * the atom is activated and the makecost and breakcost are updated.
   * 
   * @param atomIdx Index of atom to flip.
   * 
   * @return Improvement achieved by flipping the atom.
   */
  long double getImprovementByFlipping(const int& atomIdx)
  {
    if (lazy && !isActive(atomIdx))
    {
        // First flip the atom to activate it, then flip it back
      flipAtom(atomIdx);
      flipAtom(atomIdx);
    }
    long double improvement = makeCost[atomIdx] - breakCost[atomIdx];
    return improvement;
  }
  
  /**
   * If in lazy mode, an atom is activated and all clauses activated by this atom
   * are added to the state. If in eager mode, nothing happens.
   * 
   * @param atomIdx Index of atom to be activated.
   */
  void activateAtom(const int& atomIdx, const bool& ignoreActivePreds)
  {
      // Lazy version: if atom is not active, we need to activate clauses
      // and take their cost into account
    if (lazy && !isActive(atomIdx))
    {
      Predicate* p = gndPredHashArray_[atomIdx-1]->createEquivalentPredicate(domain_);
      getActiveClauses(p, newClauses_, true, ignoreActivePreds);
        // Add the clauses and preds and fill info arrays
      bool initial = false;
      addNewClauses(initial);
        // Set active status in db
      domain_->getDB()->setActiveStatus(p, true);
      activeAtoms++;
      delete p;
    }        
  }

  /**
   * Checks if an atom is active.
   * 
   * @param atomIdx Index of atom to be checked.
   * @return true, if atom is active, otherwise false.
   */  
  bool isActive(const int& atomIdx)
  {
    return domain_->getDB()->getActiveStatus(gndPredHashArray_[atomIdx-1]);
  }
  
  /**
   * Checks if an atom is active.
   * 
   * @param pred Predicate to be checked.
   * @return true, if atom is active, otherwise false.
   */  
  bool isActive(const Predicate* pred)
  {
    return domain_->getDB()->getActiveStatus(pred);
  }

  /**
   * Retrieves the occurence array of an atom.
   */
  Array<int>& getOccurenceArray(const int& idx)
  {
    return occurence[idx];
  }
  
  /**
   * Increments the number of true literals in a clause.
   */
  int incrementNumTrueLits(const int& clauseIdx)
  {
    return ++numTrueLits[clauseIdx];
  }

  /**
   * Decrements the number of true literals in a clause.
   */
  int decrementNumTrueLits(const int& clauseIdx)
  {
    return --numTrueLits[clauseIdx];
  }

  /**
   * Retrieves the cost associated with a clause.
   */
  long double getClauseCost(const int& clauseIdx)
  {
    return clauseCost[clauseIdx];
  }
  
  /**
   * Retrieves the atoms in a clauses as an Array of ints.
   */
  Array<int>& getAtomsInClause(const int& clauseIdx)
  {
    return clause[clauseIdx];
  }

  /**
   * Marks a clause as false in the state.
   */
  void addFalseClause(const int& clauseIdx)
  {
    falseClause[numFalseClauses] = clauseIdx;
    whereFalse[clauseIdx] = numFalseClauses;
    numFalseClauses++;
    costOfFalseClauses += abs(clauseCost[clauseIdx]);
  }
  
  /**
   * Unmarks a clause as false in the state.
   */
  void removeFalseClause(const int& clauseIdx)
  {
    numFalseClauses--;
    falseClause[whereFalse[clauseIdx]] = falseClause[numFalseClauses];
    whereFalse[falseClause[numFalseClauses]] = whereFalse[clauseIdx];
    costOfFalseClauses -= abs(clauseCost[clauseIdx]);
  }

  /**
   * Increases the breakcost of an atom.
   */
  void addBreakCost(const int& atomIdx, const long double& cost)
  {
    breakCost[atomIdx] += cost;
  }

  /**
   * Decreases the breakcost of an atom.
   */
  void subtractBreakCost(const int& atomIdx, const long double& cost)
  {
    breakCost[atomIdx] -= cost;
  }

  /**
   * Increases breakCost of all atoms in a given clause.
   * 
   * @param clauseIdx Index of clause whose atoms' breakCost is altered.
   * @param cost Cost to be added to atoms' breakCost.
   */
  void addBreakCostToAtomsInClause(const int& clauseIdx, const long double& cost)
  {
    register int size = clauseSize[clauseIdx];
    for (int i = 0; i < size; i++)
    {
      register int lit = clause[clauseIdx][i];
      breakCost[abs(lit)] += cost;
    }
  }

  /**
   * Decreases breakCost of all atoms in a given clause.
   * 
   * @param clauseIdx Index of clause whose atoms' breakCost is altered.
   * @param cost Cost to be subtracted from atoms' breakCost.
   */
  void subtractBreakCostFromAtomsInClause(const int& clauseIdx, const long double& cost)
  {
    register int size = clauseSize[clauseIdx];
    for (int i = 0; i < size; i++)
    {
      register int lit = clause[clauseIdx][i];
      breakCost[abs(lit)] -= cost;
    }
  }

  /**
   * Increases makeCost of an atom.
   * 
   * @param atomIdx Index of atom whose makeCost is altered.
   * @param cost Cost to be added to atom's makeCost.
   */
  void addMakeCost(const int& atomIdx, const long double& cost)
  {
    makeCost[atomIdx] += cost;
  }

  /**
   * Decreases makeCost of an atom.
   * 
   * @param atomIdx Index of atom whose makeCost is altered.
   * @param cost Cost to be subtracted from atom's makeCost.
   */
  void subtractMakeCost(const int& atomIdx, const long double& cost)
  {
    makeCost[atomIdx] -= cost;
  }

  /**
   * Increases makeCost of all atoms in a given clause.
   * 
   * @param clauseIdx Index of clause whose atoms' makeCost is altered.
   * @param cost Cost to be added to atoms' makeCost.
   */
  void addMakeCostToAtomsInClause(const int& clauseIdx, const long double& cost)
  {
    register int size = clauseSize[clauseIdx];
    for (int i = 0; i < size; i++)
    {
      register int lit = clause[clauseIdx][i];
      makeCost[abs(lit)] += cost;
    }
  }

  /**
   * Decreases makeCost of all atoms in a given clause.
   * 
   * @param clauseIdx Index of clause whose atoms' makeCost is altered.
   * @param cost Cost to be subtracted from atoms' makeCost.
   */
  void subtractMakeCostFromAtomsInClause(const int& clauseIdx, const long double& cost)
  {
    register int size = clauseSize[clauseIdx];
    for (int i = 0; i < size; i++)
    {
      register int lit = clause[clauseIdx][i];
      makeCost[abs(lit)] -= cost;
    }
  }

  /**
   * Retrieves a true literal in a clause other than the two given.
   * 
   * @param clauseIdx Index of clause from which literal is retrieved.
   * @param atomIdx1 Index of first atom excluded from search.
   * @param atomIdx2 Index of second atom excluded from search.
   * 
   * @return Index of atom found.
   */
  const int getTrueLiteralOtherThan(const int& clauseIdx,
                                    const int& atomIdx1,
                                    const int& atomIdx2)
  {
    register int size = clauseSize[clauseIdx];
    for (int i = 0; i < size; i++)
    {
      register int lit = clause[clauseIdx][i];
      register int v = abs(lit);
      if (isTrueLiteral(lit) && v != atomIdx1 && v != atomIdx2)
        return v;
    }
      // If we're here, then no other true lit exists
    assert(false);
    return -1;
  }
  
  /**
   * Checks if a literal is true (Negated and false or non-negated an true).
   */
  const bool isTrueLiteral(const int& literal)
  {
    return ((literal > 0) == atom[abs(literal)]);
  }

  /**
   * Retrieves the absolute index of the nth atom in a clause.
   */
  const int getAtomInClause(const int& atomIdxInClause, const int& clauseIdx)
  {
    return clause[clauseIdx][atomIdxInClause];
  }

  /**
   * Retrieves the absolute index of a random atom in a clause.
   */
  const int getRandomAtomInClause(const int& clauseIdx)
  {
    return clause[clauseIdx][random()%clauseSize[clauseIdx]];
  }

  /**
   * Retrieves the index of a random true literal in a clause.
   * 
   * @param clauseIdx Index of clause from which the literal is retrieved.
   * @return Index of the atom retrieved.
   */  
  const int getRandomTrueLitInClause(const int& clauseIdx)
  {
    int trueLit = random()%numTrueLits[clauseIdx];
    int whichTrueLit = 0;
    for (int i = 0; i < clauseSize[clauseIdx]; i++)
    {
      int lit = clause[clauseIdx][i];
      int atm = abs(lit);
        // True literal
      if (isTrueLiteral(lit))
        if (trueLit == whichTrueLit++)
          return atm;
    }
      // If we're here, then no other true lit exists
    assert(false);
    return -1;
  }
  
  const long double getMakeCost(const int& atomIdx)
  {
    return makeCost[atomIdx];
  }
   
  const long double getBreakCost(const int& atomIdx)
  {
    return breakCost[atomIdx];
  }
   
  const int getClauseSize(const int& clauseIdx)
  {
    return clauseSize[clauseIdx];
  }

  const int getWatch1(const int& clauseIdx)
  {
    return watch1[clauseIdx];
  }

  void setWatch1(const int& clauseIdx, const int& atomIdx)
  {
    watch1[clauseIdx] = atomIdx;
  }
  
  const int getWatch2(const int& clauseIdx)
  {
    return watch2[clauseIdx];
  }

  void setWatch2(const int& clauseIdx, const int& atomIdx)
  {
    watch2[clauseIdx] = atomIdx;
  }
  
  const bool isBlockEvidence(const int& blockIdx)
  {
    return (*blockEvidence_)[blockIdx];
  }

  const int getBlockSize(const int& blockIdx)
  {
    return (*blocks_)[blockIdx].size();
  }
  
  /** 
   * Returns the index of the block which the atom with index atomIdx
   * is in. If not in any, returns -1.
   */
  const int getBlockIdx(const int& atomIdx)
  {
    for (int i = 0; i < blocks_->size(); i++)
    {
      int blockIdx = (*blocks_)[i].find(atomIdx);
      if (blockIdx >= 0)
        return i;
    }
    return -1;
  }
  
  Array<int>& getBlockArray(const int& blockIdx)
  {
    return (*blocks_)[blockIdx];
  }

  /**
   * Returns the cost of bad clauses in the optimum state.
   */
  const long double getLowCost()
  {
    return lowCost; 
  }

  /**
   * Returns the number of bad clauses in the optimum state.
   */
  const int getLowBad()
  {
    return lowBad;
  }

  /**
   * Save current assignment as an optimum state.
   */
  void saveLowState()
  {
    for (int i = 1; i <= numAtoms; i++) lowAtom[i] = atom[i];
    lowCost = costOfFalseClauses;
    lowBad = numFalseClauses;
  }

  /**
   * Returns true if a true fixed atom is in block, otherwise false    
   */
  bool trueFixedAtomInBlock(const int blockIdx)
  {
    Array<int>& block = (*blocks_)[blockIdx];
    for (int i = 0; i < block.size(); i++)
      if (fixedAtom[block[i] + 1] > 0) return true;
    return false;
  }

  /**
   * Sets the truth values of all atoms in a block except
   * for the one given.
   * 
   * @param atomIdx Index of atom in block exempt from being set to false.
   * @param blockIdx Index of block whose atoms are set to false.
   */
  void setOthersInBlockToFalse(const int& atomIdx,
                               const int& blockIdx)
  {
    Array<int>& block = (*blocks_)[blockIdx];
    for (int i = 0; i < block.size(); i++)
    {
      if (i != atomIdx)
        setValueOfAtom(block[i] + 1, false);
    }
  }

  /**
   * Fixes an atom to a truth value. This means the atom can not change
   * its truth values again. If the atom has already been fixed to the
   * opposite value, then we have a contradiction and the program terminates.
   * 
   * @param atomIdx Index of atom to be fixed.
   * @param value Truth value to which the atom is fixed.
   */
  void fixAtom(const int& atomIdx, const bool& value)
  {
    assert(atomIdx > 0);
      // If already fixed to opp. sense, then contradiction
    if ((fixedAtom[atomIdx] == 1 && value == false) ||
        (fixedAtom[atomIdx] == -1 && value == true))
    {
      cout << "Contradiction: Tried to fix atom " << atomIdx <<
      " to true and false ... exiting." << endl;
      exit(0);
    }
    setValueOfAtom(atomIdx, value);
    fixedAtom[atomIdx] = (value) ? 1 : -1;
  }

  /**
   * Simplifies a clause using atoms which have been fixed. If clause is
   * satisfied from the fixed atoms, this is marked in isSatisfied and an empty
   * array is returned. If clause is empty and not satisfied, then a contradiction
   * has occured. Otherwise, the simplified clause is returned.
   * 
   * Returned array should be deleted.
   * 
   * @param clauseIdx Index of the clause to be simplified
   * @return Simplified clause
   */
  Array<int>* simplifyClauseFromFixedAtoms(const int& clauseIdx)
  {
    Array<int>* returnArray = new Array<int>;
      // If already satisfied from fixed atoms, then return empty array
    if (isSatisfied[clauseIdx]) return returnArray;

      // Keeps track of pos. clause being satisfied or 
      // neg. clause being unsatisfied due to fixed atoms
    bool isGood = (clauseCost[clauseIdx] > 0) ? false : true;
      // Keeps track of all atoms being fixed to false in a pos. clause
    bool allFalseAtoms = (clauseCost[clauseIdx] > 0) ? true : false;
      // Check each literal in clause
    for (int i = 0; i < clauseSize[clauseIdx]; i++)
    {
      int lit = clause[clauseIdx][i];
      int fixedValue = fixedAtom[abs(lit)];

      if (clauseCost[clauseIdx] > 0)
      { // Pos. clause: check if clause is satisfied
        if ((fixedValue == 1 && lit > 0) ||
            (fixedValue == -1 && lit < 0))
        { // True fixed lit
          isGood = true;
          allFalseAtoms = false;
          returnArray->clear();
          break;
        }
        else if (fixedValue == 0)
        { // Lit not fixed
          allFalseAtoms = false;
          returnArray->append(lit);
        }
      }
      else
      { // Neg. clause:
        assert(clauseCost[i] < 0);
        if ((fixedValue == 1 && lit > 0) ||
            (fixedValue == -1 && lit < 0))
        { // True fixed lit
          cout << "Contradiction: Tried to fix atom " << abs(lit) <<
          " to true in a negative clause ... exiting." << endl;
          exit(0);
        }
        else
        { // False fixed lit or non-fixed lit
          returnArray->append(lit);
            // Non-fixed lit
          if (fixedValue == 0) isGood = false;          
        }
      }
    }
    if (allFalseAtoms)
    {
      cout << "Contradiction: All atoms in clause " << clauseIdx <<
      " fixed to false ... exiting." << endl;
      exit(0);        
    }
    if (isGood) isSatisfied[clauseIdx] = true;
    return returnArray;
  }
  
  void setLazy(const bool& l) { lazy = l; }
  bool getLazy() { return lazy; }

  /**
   * Prints the probabilities of each predicate to a stream. For SAT-solvers
   * this is the best state found.
   */
  void printProbabilities(ostream& out) const
  {
    for (int i = 0; i < numAtoms; i++)
    {
      (*gndPreds_)[i]->print(out, domain_);
      out << " " << lowAtom[i + 1] << endl;
    }
  }

  /**
   * Prints the predicates set to true in the best state to a stream.
   */
  void printTruePreds(ostream& out) const
  {
    for (int i = 0; i < numAtoms; i++)
    {
      if (lowAtom[i + 1]) 
      {
        (*gndPreds_)[i]->print(out, domain_);
        out << endl;
      }
    }
  }

  ////////////// BEGIN: Lazy Functions //////////////
 
  /** 
   * Gets clauses and weights activated by the predicate inputPred,
   * if active is true. If false, inactive clauses (and their weights)
   * containing inputPred are retrieved. If inputPred is NULL, then all
   * active (or inactive) clauses and their weights are retrieved.
   */
  void getActiveClauses(Predicate *inputPred,
                        Array<GroundClause*> &activeClauses,
                        bool const & active,
                        bool const & ignoreActivePreds)
  {
    Clause *fclause;
    //IntClause *intClause;
    GroundClause* newClause;
    int clauseCnt;
    //IntClauseHashArray clauseHashArray;
    GroundClauseHashArray clauseHashArray;

    //Array<IntClause *>* intClauses = new Array<IntClause *>; 
    Array<GroundClause*>* newClauses = new Array<GroundClause*>; 
  
    const Array<IndexClause*>* indexClauses = NULL;
      
      // inputPred is null: all active clauses should be retrieved
    if (inputPred == NULL)
    {
      clauseCnt = mln_->getNumClauses();
    }
      // Otherwise, look at all first order clauses containing the pred
    else
    {
      if (domain_->getDB()->getDeactivatedStatus(inputPred)) return;
      int predId = inputPred->getId();
      indexClauses = mln_->getClausesContainingPred(predId);
      clauseCnt = indexClauses->size();
    }

    int clauseno = 0;
    while (clauseno < clauseCnt)
    {
      if (inputPred)
        fclause = (Clause *) (*indexClauses)[clauseno]->clause;           
      else
        fclause = (Clause *) mln_->getClause(clauseno);
    
      long double wt = fclause->getWt();
      //intClauses->clear();
      newClauses->clear();

      fclause->getActiveClauses(inputPred, domain_, newClauses,
                                NULL, ignoreActivePreds);
//cout << "newClauses size " << newClauses->size() << endl;
      for (int i = 0; i < newClauses->size(); i++)
      {
        newClause = (*newClauses)[i];
        int pos = clauseHashArray.find(newClause);
        if (pos >= 0)
        {
          clauseHashArray[pos]->addWt(wt);
          //intClause->deleteIntPredicates();
          //delete intClause;
          newClause->deleteAllGndPreds();
          delete newClause;
          continue;
        }
           
        newClause->setWt(wt);
        clauseHashArray.append(newClause);
      }
      clauseno++; 
    } //while (clauseno < clauseCnt)
      
    for (int i = 0; i < clauseHashArray.size(); i++)
    {
      newClause = clauseHashArray[i];
      activeClauses.append(newClause);
      
        // Add new gndPreds to state
      int numPreds = newClause->getNumGroundPredicates();
  
        // For each predicate in new clause, add it to state if not yet there
      for (int i = 0; i < numPreds; i++)
      {
        GroundPredicate* pred = (GroundPredicate*)newClause->getGroundPredicate(i);
        int index = gndPredHashArray_.find(pred);
        if (index < 0)
          index = gndPredHashArray_.append(pred);
        newClause->setGroundPredicateIndex(i, index);
      }
    }
    delete newClauses;
  }

  /**
   * Get all the active clauses in the database.
   */
  void getActiveClauses(Array<GroundClause*> &allClauses)
  {
    bool ignoreActivePreds = false;
    getActiveClauses(NULL, allClauses, true, ignoreActivePreds);
  }
  
  int getNumActiveAtoms()
  {
    return activeAtoms; 
  }

  /**
   * Initializes the block structures for the lazy version. All the
   * non-evidence predicates in blocks in the domain are copied into the
   * blocks of the state.
   */
  void initLazyBlocks()
  {
    blocks_ = new Array<Array<int> >;
    blocks_->growToSize(domain_->getNumPredBlocks());
    blockEvidence_ = new Array<bool>(*(domain_->getBlockEvidenceArray()));

    const Array<Array<Predicate*>*>* blocks = domain_->getPredBlocks();
    for (int i = 0; i < blocks->size(); i++)
    {
      Array<Predicate*>* block = (*blocks)[i];
      for (int j = 0; j < block->size(); j++)
      {
          // Add all non-evid preds in blocks to the state
        Predicate* pred = (*block)[j];
        if (domain_->getDB()->getEvidenceStatus(pred))
          continue;
        GroundPredicate* groundPred = new GroundPredicate(pred);

          // Add pred if not yet present, otherwise delete it
        int index = gndPredHashArray_.find(groundPred);
        if (index < 0)
          index = gndPredHashArray_.append(groundPred);
        else
          delete groundPred;

        (*blocks_)[i].append(index);
      }
    }
  }

  ////////////// END: Lazy Functions //////////////

  
 private:
 
  /**
   * Sets the hard clause weight to the sum of all soft clause weights in
   * the domain (not in the MRF) plus a constant (10). This is then the hard
   * clause weight when no soft clauses are present.
   */
  void setHardClauseWeight()
  {
       // Soft weights are summed up to determine hard weight
    long double sumSoftWts = 0.0;
      // Determine hard clause weight
    int clauseCnt = mln_->getNumClauses();    
      // Sum up the soft weights of all grounded clauses
    for (int i = 0; i < clauseCnt; i++)
    {
      Clause* fclause = (Clause *) mln_->getClause(i);
        // Skip hard clauses
      if (fclause->isHardClause()) continue;
        // Weight could be negative
      long double wt = abs(fclause->getWt());
      long double numGndings = fclause->getNumGroundings(domain_);
      sumSoftWts += wt*numGndings;
    }
    assert(sumSoftWts >= 0);
      // Add constant so weight isn't zero if no soft clauses present
    hardWt = sumSoftWts + 10.0;
    cout << "Set hard weight to " << hardWt << endl;
  }

 private:

    // If true, this is a lazy variable state, else eager.
  bool lazy;

    // Weight used for hard clauses (sum of soft weights + constant)
  long double hardWt;
  
    // mln and domain are used to build MRF in eager state and to
    // retrieve active atoms in lazy state.
  MLN* mln_;
  Domain* domain_;

    // Eager version: Pointer to gndPreds_ and gndClauses_ in MRF
    // Lazy version: Holds active atoms and clauses
  Array<GroundPredicate*>* gndPreds_;
  Array<GroundClause*>* gndClauses_;
  //GroundPredicateHashArray gndPreds_;
  //GroundClauseHashArray gndClauses_;

    ////////// BEGIN: Information used only by lazy version //////////
    // Number of distinct atoms in the first set of unsatisfied clauses
  int baseNumAtoms;
    // If true, atoms are not deactivated when mem. is full
  bool noApprox;
    // Indicates whether deactivation of atoms has taken place yet
  bool haveDeactivated;
    // Max. amount of memory to use
  int memLimit;
    // Max. amount of clauses memory can hold
  int clauseLimit;
    ////////// END: Information used only by lazy version //////////


    ////////// BEGIN: Information used only by eager version //////////
    // MRF is used with eager states. If lazy, this stays NULL.
  MRF* mrf_;
    ////////// END: Information used only by eager version //////////

    // Holds the new active clauses
  Array<GroundClause*> newClauses_;
    // Holds the ground clauses in a hash array.
    // Fast access is needed for comparison when activating clauses.
  //GroundClauseHashArray gndClauseHashArray_;
    // Holds the new gnd preds
  Array<GroundPredicate*> newPreds_;
    // Holds the ground predicates in a hash array.
    // Fast access is needed for comparing preds when activating clauses.
  GroundPredicateHashArray gndPredHashArray_;

    // Total no. of atoms
  int numAtoms;
    // Total no. of clauses
  int numClauses;
  
    // Clauses to be satisfied
    // Indexed as clause[clause_num][literal_num]
  Array<Array<int> > clause;
    // Cost of each clause (can be negative)
  Array<long double> clauseCost;
    // Highest cost of false clause
  long double highestCost;
    // If true, more than one clause has highest cost
  bool eqHighest;
    // Number of clauses with highest cost
  int numHighest;
    // Length of each clause
  Array<int> clauseSize;
    // Clauses which are pos. and unsatisfied or neg. and satisfied
  Array<int> falseClause;
    // Where each clause is listed in falseClause
  Array<int> whereFalse;
    // Number of true literals in each clause
  Array<int> numTrueLits;
    // watch1[c] contains the id of the first atom which c is watching
  Array<int> watch1;
    // watch2[c] contains the id of the second atom which c is watching
  Array<int> watch2;
    // Which clauses are satisfied by fixed atoms
  Array<bool> isSatisfied;

    // Holds the index of clauses in which each literal occurs
    // Indexed as occurence[2*abs(lit) - (lit > 0)][occurence_num]
  Array<Array<int> > occurence;

    // Current assigment of atoms
  Array<bool> atom;
  //int* solution;
    // Cost of clauses which would become satisfied by flipping each atom
  Array<long double> makeCost;
    // Cost of clauses which would become unsatisfied by flipping each atom
  Array<long double> breakCost;
    // Indicates if an atom is fixed to a value (0 = not fixed, -1 = false, 1 = true)
  Array<int> fixedAtom;

    // Assigment of atoms producing lowest cost so far
  Array<bool> lowAtom;
    // Cost of false clauses in the currently best state
  long double lowCost;
    // Number of false clauses in the currently best state
  int lowBad;

    // Current no. of unsatisfied clauses
  int numFalseClauses;
    // Cost associated with the number of false clauses
  long double costOfFalseClauses;
  
    // For block inference: blocks_ and blockEvidence_
    // All atom indices in (*blocks_)[i] are to be treated as in one block
  Array<Array<int> >* blocks_;
    // (*blockEvidence_)[i] states whether block i has true evidence and
    // thus all should be false
  Array<bool >* blockEvidence_;

    // Number of active atoms in state.  
  int activeAtoms;
  
};

#endif /*VARIABLESTATE_H_*/
